# -*- coding: utf-8 -*-


class LimsException(Exception):
    pass


class MissingLimsDataException(LimsException, KeyError):
    pass


class LimsSampleNotFoundError(LimsException):
    pass


class LimsCaseIdNotFoundError(LimsException):
    pass


class LimsSampleCancelled(LimsException):
    pass


class MultipleSamplesError(LimsException):
    pass



# -*- coding: utf-8 -*-
from cglims import pedigree


def test_sequencing_type():
    # GIVEN a whole genome application tag
    app_tag = 'WGSPCFC030'
    # WHEN assessing the sequencing type
    seq_type = pedigree.sequencing_type(app_tag)
    # THEN it should come back as 'wgs'
    assert seq_type == 'wgs'



# -*- coding: utf-8 -*-
import datetime
import json

from cglims.api import ClinicalLims


def connect_api(config):
    """Connect and return API reference."""
    api = ClinicalLims(config['host'], config['username'], config['password'])
    return api


def json_serial(obj):
    """JSON serializer for objects not serializable by default json code"""
    if isinstance(obj, datetime.datetime):
        serial = obj.isoformat()
        return serial
    raise TypeError('Type not serializable')


def jsonify(data, pretty=False):
    """Serialize Model to JSON."""
    kwargs = dict(indent=4, sort_keys=True) if pretty else dict()
    return json.dumps(data, default=json_serial, **kwargs)



# -*- coding: utf-8 -*-
import io
import copy
import csv

PHENOTYPE_MAP = dict(Affected='2', Unaffected='1', unknown='0')
SEX_MAP = dict(M='1', F='2')


class UnknownSequencingTypeError(Exception):
    pass


def get_sampleid(lims_sample, key='Clinical Genomics ID'):
    """Get the expected LIMS or Clinical Genomics ID."""
    return lims_sample.udf.get(key) or lims_sample.id


def sequencing_type(app_tag):
    """Parse application type to figure out type of sequencing."""
    if app_tag.startswith('WG'):
        return 'wgs'
    elif app_tag.startswith('EX'):
        return 'wes'
    else:
        raise UnknownSequencingTypeError(app_tag)


def make_pedigree(api, customer, family_id, internalize=True):
    """String together all individual steps to create a pedigree."""
    # find sample in lims
    lims_samples = api.case(customer, family_id)
    # extract information about samples
    samples = [convert_sample(lims_sample) for lims_sample in lims_samples
               if lims_sample.udf.get('cancelled') != 'yes']
    if internalize:
        samples = internalize_ids(samples)
    # serialize to CSV PED file
    rows = serialize(samples)
    return "#{}".format(rows)


def convert_sample(lims_sample):
    """Extract information from LIMS samples that relate to pedigree."""
    app_tag = lims_sample.udf['Sequencing Analysis']
    affection_status = lims_sample.udf['Status']
    sex_letter = lims_sample.udf['Gender']
    data = {
        'Family ID': lims_sample.udf['familyID'],
        'Individual ID': lims_sample.name,
        'Paternal ID': lims_sample.udf.get('fatherID', '0'),
        'Maternal ID': lims_sample.udf.get('motherID', '0'),
        'Sex': SEX_MAP[sex_letter],
        'Phenotype': PHENOTYPE_MAP[affection_status],
        'Clinical_db': lims_sample.udf['Gene List'],
        'Sequencing_type': sequencing_type(app_tag),
        'internal_id': get_sampleid(lims_sample),
    }
    return data


def internalize_ids(samples):
    """Replace customer sample ids with internal ids."""
    sample_map = {sample['Individual ID']: sample for sample in samples}
    for external_id, sample in sample_map.items():
        new_sample = copy.deepcopy(sample)
        new_sample['Individual ID'] = sample['internal_id']
        new_sample['display_name'] = external_id

        parent_fields = ['Paternal ID', 'Maternal ID']
        for parent_field in parent_fields:
            parent_id = new_sample[parent_field]
            if parent_id != '0':
                new_sample[parent_field] = sample_map[parent_id]['internal_id']
        yield new_sample


def serialize(sample_dicts, headers=None, extra=None):
    """Serialize a list of samples into PED output to a file.

    Args:
        sample_dicts (List[dict]): list of dicts with values
        headers (Optional[list]): order of headers to output
    """
    output = io.BytesIO()
    headers = headers or ['Family ID', 'Individual ID', 'Paternal ID',
                          'Maternal ID', 'Sex', 'Phenotype']
    extra = extra or ['Clinical_db', 'display_name', 'Sequencing_type']
    all_headers = headers + extra
    dict_writer = csv.DictWriter(output, all_headers, extrasaction='ignore',
                                 dialect='excel-tab', lineterminator='\n')
    dict_writer.writeheader()
    dict_writer.writerows(sample_dicts)
    return output.getvalue()



# -*- coding: utf-8 -*-
from datetime import datetime
from copy import deepcopy

import click

from cglims.pedigree import make_pedigree
from .utils import connect_api, jsonify


@click.command()
@click.argument('customer')
@click.argument('family')
@click.pass_context
def pedigree(context, customer, family):
    """Create pedigree from LIMS."""
    api = connect_api(context.obj)
    content = make_pedigree(api, customer, family)
    click.echo(content, nl=False)


@click.command()
@click.option('-p', '--pretty', is_flag=True, help='pretty print JSON')
@click.option('-f', '--field', help='print specific field only')
@click.argument('identifier')
@click.pass_context
def get(context, pretty, field, identifier):
    """Get information from LIMS: either sample or family samples."""
    api = connect_api(context.obj)
    if identifier.startswith('cust'):
        # look up samples in a case
        samples = api.case(*identifier.split('-', 1))
    else:
        # look up a single sample
        is_cgid = True if identifier[0].isdigit() else False
        samples = [api.sample(identifier, is_cgid=is_cgid)]

    for sample in samples:
        fields = deepcopy(sample.udf._lookup)
        fields['id'] = sample.id
        fields['name'] = sample.name
        date_parts = map(int, sample.date_received.split('-'))
        fields['date_received'] = datetime(*date_parts)
        fields['project_name'] = sample.project.name

        if field:
            click.echo("{}: {}".format(sample.id, fields.get(field, 'N/A')))
        else:
            click.echo(jsonify(fields, pretty=pretty))


@click.command()
@click.argument('lims_id')
@click.argument('field_key')
@click.argument('new_value')
@click.pass_context
def update(context, lims_id, field_key, new_value):
    """Update a UDF for a sample."""
    api = connect_api(context.obj)
    lims_sample = api.sample(lims_id)
    old_value = lims_sample.udf.get(field_key, 'N/A')
    click.echo("about to update sample: {}".format(lims_sample.id))
    message_tmlt = "are you sure you want to change '{}': '{}' -> '{}'"
    if click.confirm(message_tmlt.format(field_key, old_value, new_value)):
        lims_sample.udf[field_key] = new_value
        lims_sample.put()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import codecs
import sys
from setuptools import find_packages, setup
from setuptools.command.test import test as TestCommand


def parse_reqs(req_path='./requirements.txt'):
    """Recursively parse requirements from nested pip files."""
    install_requires = []
    with codecs.open(req_path, 'r') as handle:
        # remove comments and empty lines
        lines = (line.strip() for line in handle
                 if line.strip() and not line.startswith('#'))
        for line in lines:
            # check for nested requirements files
            if line.startswith('-r'):
                # recursively call this function
                install_requires += parse_reqs(req_path=line[3:])
            else:
                # add the line as a new requirement
                install_requires.append(line)
    return install_requires


# This is a plug-in for setuptools that will invoke py.test
# when you run python setup.py test
class PyTest(TestCommand):

    """Set up the py.test test runner."""

    def finalize_options(self):
        """Set options for the command line."""
        TestCommand.finalize_options(self)
        self.test_args = ['-v']
        self.test_suite = True

    def run_tests(self):
        """Execute the test runner command."""
        # Import here, because outside the required eggs aren't loaded yet
        import pytest
        sys.exit(pytest.main(self.test_args))

setup(
    name='cglims',
    version='1.0.0-beta2',
    description='Microservice to interface with LIMS',
    author='Robin Andeer',
    author_email='robin.andeer@scilifelab.se',
    packages=find_packages(exclude=('tests*', 'docs', 'examples')),
    include_package_data=True,
    install_requires=parse_reqs(),
    cmdclass=dict(test=PyTest),
    zip_safe=False,
    keywords='pytest cli',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
    ],
    platforms='any',
    license='BSD License',
    entry_points={
        'console_scripts': ['cglims = cglims.cli:root'],
        'cglims.subcommands.1': [
            'pedigree = cglims.cli.commands:pedigree',
            'get = cglims.cli.commands:get',
            'update = cglims.cli.commands:update',
        ],
    },
)



# -*- coding: utf-8 -*-
from genologics.entities import Sample
from genologics.lims import Lims

from cglims.exc import MultipleSamplesError


class ClinicalLims(Lims):

    def case(self, customer, family_id):
        filters = {'customer': customer, 'familyID': family_id}
        samples = self.get_samples(udf=filters)
        return samples

    def sample(self, lims_id, is_cgid=False):
        """Get a unique sample from LIMS."""
        if is_cgid:
            udf_key = 'Clinical Genomics ID'
            lims_samples = self.get_samples(udf={udf_key: lims_id})
            if len(lims_samples) == 1:
                return lims_samples[0]
            elif len(lims_samples) > 1:
                matches_str = ', '.join(sample.id for sample in lims_samples)
                message = "'{}' matches: {}".format(lims_id, matches_str)
                raise MultipleSamplesError(message)
            else:
                # no matching samples
                return None
        else:
            lims_sample = Sample(self, id=lims_id)
        return lims_sample


def parse_udfs(udfs):
    """Parse raw UDF values for a sample."""
    pass
